/* Mehrspieler-Rasterbike fr via Netzwerk: Autopilot
   Der Autopilot simuliert ein Spieler (Computergegner)
    2000 by Andreas Meile, CH-8242 Hofen SH */

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>

/* Spielstandbeschreibung */

#define STD_IN            0  /* Standard-Eingabegert: Nur bei Master */
#define RCV_BUF          50  /* Empfangspuffer */
#define STD_PORT_NR   11971  /* TCP/IP-Portnummer, wo der Server horcht */
#define MAX_SPIELER      10  /* maximale Anzahl Spieler */
#define FELD_GROESSE     50  /* Grsse des Spielfeldes */
#define TEL_PUF          20  /* Grsse Telegramm-Empfangsliste */

struct telegramm_puffer {
  char txt[RCV_BUF];   /* Telegramminhalt */
  int lg;              /* Lnge */
};

struct spieler {
  int x, y,   /* Position des Spielers im Feld */
      lebend; /* Flag, ob immer noch lebend */
};

const int sx[] = { 1,  0, -1, 0 };
const int sy[] = { 0, -1,  0, 1 };

struct telegramm_puffer tpuf[TEL_PUF];

struct spieler sp[MAX_SPIELER];

int spielfeld[FELD_GROESSE + 2][FELD_GROESSE + 2],  /* Spielfeldzustand */
    fd_max,              /* fr select() bentigt */
    n_tel,               /* nchstes zu bearbeitendes Telegramm */
    nf_tel,              /* nchster freier Platz */
    sd_srv,              /* Socket-Deskriptor an Server */
    anz_by;              /* Anzahl Bytes im Puffer */

unsigned char empf_puf[RCV_BUF]; /* fr Socket-Empfang */

struct timeval n_ereig;  /* Stattfindung nchstes Ereignis */
struct timezone zz;      /* nicht verwertet */

char *prg_name;

/* Allgemeine Fehlerbehandlung */
void fehler(char *grund) {
  fprintf(stderr, "Fehler bei %s\n", grund);
  perror(prg_name);
  exit(1);
}

/* Zeitarithmetik */

void inkremiere_zeit(struct timeval *zw, long diff) {
  zw->tv_usec += diff;
  zw->tv_sec += zw->tv_usec / 1000000;
  zw->tv_usec %= 1000000;
}

void zeit_diff(const struct timeval *abs1,
               const struct timeval *abs2, struct timeval *erg) {
  erg->tv_sec = abs1->tv_sec - abs2->tv_sec;
  erg->tv_usec = abs1->tv_usec - abs2->tv_usec;
  if(erg->tv_usec < 0) {
    erg->tv_sec--;
    erg->tv_usec+=1000000;
  }
}

void initialisiere_spielfeld(void) {
  int i, j;
  for(i = 0; i < FELD_GROESSE + 2; i++) {
    spielfeld[i][0] = 1;
    spielfeld[i][FELD_GROESSE + 1] = 1;
  }
  for(i = 1; i <= FELD_GROESSE; i++) {
    spielfeld[0][i] = 1;
    for(j = 1; j <= FELD_GROESSE; j++)
      spielfeld[j][i] = 0;
    spielfeld[FELD_GROESSE + 1][i] = 1;
  }
}

/* Autopilotseitiges Socket-Schliessen: Empfangspuffer leerlesen */
void schliesse(void) {
  char dummy[20];
  int fcntl_flags, a_by;
  extern int errno;

  if(sd_srv == -1)
    fprintf(stderr, "Warnung: Zweites Schliessen ignoriert\n");
  else {
    /* Empfangene Bytes leerlesen */
    fcntl_flags = fcntl(sd_srv, F_GETFL);
    if(fcntl_flags == -1)
      fehler("Flags lesen");
    fcntl_flags |= O_NONBLOCK;
    if(fcntl(sd_srv, F_SETFL, fcntl_flags) != 0)
      fehler("Flags setzen");
    do {
      a_by = read(sd_srv, dummy, sizeof(dummy));
    } while(a_by > 0);
    if(a_by == -1 && errno != EWOULDBLOCK)
      fehler("Leerlesen");
    /* Puffer jetzt leer, es kann geschlossen werden */
    if(close(sd_srv) != 0)
      fehler("Schliessen (Abweisen)");
  }
}

void schreibe_telegramm(char *telegramm, int telegr_l) {
  int i, j;
  unsigned char buf[20];

  if(sd_srv == -1)
    fprintf(stderr, "Warnung: berflssiges Telegramm ignoriert\n");
  else {
    for(i=0; i<telegr_l; i++) {
      buf[i] = telegramm[i];
    }
    buf[telegr_l - 1] |= 0x80;  /* Kennung fr Telegrammende */
    i=0;
    while(i < telegr_l) {
      j = write(sd_srv, buf + i, sizeof(char) * (telegr_l - i));
      if(j < 0)
        fehler("berschreiben des Sockets");
      i += j;    
    }
  }
}

/* Socket von Server lesen und dabei erhaltenes Telegramm zurckliefern */
int lies_telegramm(int modus) {
  fd_set warte_liste;    /* fr gleichzeitiges Warten */
  int h, h2, i, temp_sd, n_sel;
  char dummy[20];
  size_t a_by;
  struct timeval akt_z, wart_z;

  n_sel = 1;
  while(nf_tel == n_tel && n_sel != 0) {
    FD_ZERO(&warte_liste);
    if(n_ereig.tv_usec != -1) {
      if(gettimeofday(&akt_z, &zz) != 0)
        fehler("Zeit ermitteln");
      zeit_diff(&n_ereig, &akt_z, &wart_z);
    }
    if(n_ereig.tv_usec == -1 || wart_z.tv_sec > 0 || wart_z.tv_sec == 0 &&
    wart_z.tv_usec > 0) {
      if(modus == 1)  /* Nur, wenn Masterstatus, dann Tastendruck abwarten */
        FD_SET(STD_IN, &warte_liste);
      FD_SET(sd_srv, &warte_liste);
      n_sel = select(fd_max, &warte_liste, NULL, NULL, n_ereig.tv_usec != -1 ?
      &wart_z : NULL);
    } else
      n_sel = 0;  /* simuliere select() mit Time Out */
    if(n_sel < 0)
      fehler("Abwarten");
    if(FD_ISSET(STD_IN, &warte_liste)) {
      a_by = read(STD_IN, dummy, sizeof(dummy));
      if(a_by < 0)
        fehler("Tastatur lesen");
      else if(a_by = 0)
        fehler("Unerwartetes Dateiende von Tastatur");
      else if(modus != 1)
        fprintf(stderr, "Keine Master-Tastenbesttigung ntig\n");
      else
        n_sel = 0;  /* Simuliere Time Out */
    }
    if(FD_ISSET(sd_srv, &warte_liste)) {
      a_by = read(sd_srv, empf_puf + anz_by, sizeof(empf_puf) - anz_by);
      if(a_by < 0)
        fehler("Telegramm lesen");
      if(a_by == 0) {
        /* Server hat sich getrennt */
        if(modus != 2)
          fprintf(stderr, "Server hat sich unerwartet getrennt\n");
        else
          printf("Serverseitige Beendung\n");
        if(close(sd_srv) != 0)
          fehler("Schliessen (Server)");
        if(modus != 2)
          exit(1);
        sd_srv = -1;
        n_sel = 0; /* Simuliere Time Out */
      }
      /* Scannen nach Telegrammende */
      h = anz_by + a_by;
      h2 = 0;
      for(i = anz_by; i < h; i++)
        if(empf_puf[i] & 0x80) {
          /* Vollstndiges Telegramm gefunden => in Liste stellen */
          tpuf[nf_tel].lg = i + 1 - h2;
          memcpy(tpuf[nf_tel].txt, empf_puf + h2, tpuf[nf_tel].lg);
          tpuf[nf_tel].txt[i - h2] &= 0x7f;
          h2 = i + 1;
          nf_tel = (nf_tel + 1) % TEL_PUF;
          if(nf_tel == n_tel)
            fprintf(stderr,
            "WARNUNG: Telegramm-Pufferberlauf ist aufgetreten\n");
        }
      anz_by = h - h2;
      memmove(empf_puf, empf_puf + h2, h - h2);
    }
    if(anz_by == RCV_BUF) {
      fprintf(stderr, "Warnung: berlanges Telegramm erhalten => verworfen\n");
      anz_by = 0;
    }
  }
  if(n_sel != 0) {
    i = n_tel;
    n_tel = (n_tel + 1) % TEL_PUF;
  } else
    i = -1;   /* kein Telegramm erhalten, weil Anmelde-Timeout oder
                 Tastendruck */
  return i;
}

int main(int argc, char *argv[]) {
  int tel,        /* Telegramm */
      n_spieler,  /* Anzahl teilnehmende Spieler */
      eig_s,      /* Spieler, den der Autopilot ist */
      eig_r,      /* Richtung vom Autopilot */
      eig_rv,     /* vorgeschlagene Richtung */
      strat,      /* Strategie */
      zust,       /* aktueller Zustand */
      s_kehr,     /* Nur Strat. 2 & 3: Anz. Felder bis zum nchsten Abbiegen */
      port_nr,    /* effektive TCP/IP-Portnummer */
      i, j, s;    /* Hilfsvariablen */

  char r_tel;     /* Puffer fr Telegramm */

  struct timeval zufallsg_init;
  struct hostent *host_eintr;
  struct sockaddr_in srv_sock;

  printf("Rasterbike-Autopilot V1.01  2000 by Andreas Meile, CH-8242 Hofen SH\n");
  prg_name = argv[0];

  switch(argc) {
    case 3:
      port_nr = STD_PORT_NR;
      break;
    case 4:
      if(sscanf(argv[3], "%d", &port_nr) == 1)
        break;
    default:
      fprintf(stderr, "Aufrufformat: %s <Strategie> <Server> [ <Port-Nummer> ]\n"
                      "wobei fr Strategie:\n"
                      "1=primitives Ausweichen\n"
                      "2=Ausweichen & pltzliches Abbiegen\n"
                      "3=Langsamer Nachlaufer\n"
                      "4=Spieler nachhetzen\n"
                      "5=Zufalls-Herumtorkler\n"
                      "6=An-der-Wand-Schleichtaktik\n", argv[0]);
      exit(1);
  }

  if(sscanf(argv[1], "%d", &strat) != 1 || strat < 1 || strat > 6) {
    fprintf(stderr, "Strategie muss zwischen 1 und 6 liegen\n");
    exit(1);
  }

  initialisiere_spielfeld();
  n_tel = 0;   /* Zu Beginn leere Telegrammliste */
  nf_tel = 0;
  n_spieler = 0;

  /* Zufallsgenerator initialisieren */
  if(gettimeofday(&zufallsg_init, &zz) != 0)
    fehler("Zeit ermitteln");
  srand(zufallsg_init.tv_sec ^ zufallsg_init.tv_usec);

  /* Socket erzeugen */
  sd_srv = socket(AF_INET, SOCK_STREAM, 0);
  if(sd_srv == -1)
    fehler("Socket erzeugen");
  fd_max = sd_srv + 1;

  /* Hostname auflsen */
  host_eintr = gethostbyname(argv[2]);
  if(host_eintr == NULL) {
    herror("connect");
    fehler("Hostnamenauflsung");
  }

  /* Verbinden */
  memset(&srv_sock, 0, sizeof(srv_sock));
  memcpy(&srv_sock.sin_addr, host_eintr->h_addr, host_eintr->h_length);
  srv_sock.sin_family = AF_INET;
  srv_sock.sin_port   = htons(port_nr);
  if (connect(sd_srv, (struct sockaddr *)&srv_sock, sizeof(srv_sock)) != 0)
    fehler("Verbinden");

  anz_by = 0;
  printf("Verbindung mit Server aufgenommen, Handshake luft...\n");

  /* Handshake-Phase */
  schreibe_telegramm("Raster", 6);

  if(gettimeofday(&n_ereig, &zz) != 0)
    fehler("Zeit ermitteln");
  inkremiere_zeit(&n_ereig, 40000000);
  i = lies_telegramm(0);
  if(i == -1) {
    schliesse();
    fprintf(stderr, "Zeitberlauf - kein Rasterbike-Server\n");
    exit(1);
  }
  if(tpuf[i].lg != 5 || strncmp(tpuf[i].txt, "Bike", 4)) {
    schliesse();
    fprintf(stderr, "Falsche Rckantwort - kein Rasterbike-Server\n");
    exit(1);
  }
  eig_s = tpuf[i].txt[4];
  printf("Handshake beendet - bernehme Rolle von Spieler %d\n", eig_s+1);

  /* Einladungsphase */
  n_ereig.tv_usec = -1;  /* keine Uhr */
  printf("Einladungsphase luft...\n");
  zust = 0;
  while(zust != 2) {
    i = lies_telegramm(zust);
    if(i == -1) {
      printf("Einladungsphase mit Taste beendet\n");
      schreibe_telegramm("EI", 2);
    } else {
      if(tpuf[i].lg == 2) {
        if(!strncmp(tpuf[i].txt, "NY", 2))
          printf("Noch nicht alle Spieler sind bereit - "
                 "Tastendruck wiederholen\n");
        else if(!strncmp(tpuf[i].txt, "MA", 2)) {
          printf("Rolle vom Master erhalten - "
                 "Einladungsphase mit Tastendruck beenden!\n");
          zust = 1;
        } else if(!strncmp(tpuf[i].txt, "PP", 2))
          zust = 2;
        else
          fprintf(stderr, "Protokollfehler - PP erwartet\n");
      } else if(tpuf[i].txt[0] == 'P' && tpuf[i].lg % 3 == 1) {
        j = tpuf[i].lg;
        while(j > 1) {
          j -= 3;
          s = tpuf[i].txt[j+2] >> 2;
          while(n_spieler <= s) {
            sp[n_spieler].lebend = 0;
            n_spieler++;
          }
          sp[s].x = tpuf[i].txt[j];
          sp[s].y = tpuf[i].txt[j+1];
          if(s == eig_s)
            eig_r = tpuf[i].txt[j+2] & 0x3;
          sp[s].lebend = 1;
          spielfeld[sp[s].x][sp[s].y] = 1;
        }
      } else if(tpuf[i].txt[0] == 'X' && tpuf[i].lg == 3)
        sp[tpuf[i].txt[2] >> 2].lebend = 0;
      else
        fprintf(stderr, "Nicht verstandenes Telegramm erhalten - PP\n");
    } 
  }
  printf("Einladungsphase beendet - Spielerliste: ");
  for(i=0; i<n_spieler; i++)
    putchar(sp[i].lebend ? i == eig_s ? 'E' : 'G' : '_');
  printf("\nLegende: E=eigener Spieler, G=Gegner, _=Netzwerkverb. verloren\n");

  /* Bereitmachungsphase */
  printf("Beginn Bereitmachungsphase\n");
  schreibe_telegramm("R", 1);

  while(zust != 3) {
    i = lies_telegramm(0);
    if(tpuf[i].lg == 2 && tpuf[i].txt[0] == 'A') {
      switch(tpuf[i].txt[1]) {
        case '0':
          printf("Ende Bereitmachung\n");
        break;
        case '1':
        break;
        case '2':
          zust = 3;
        break;
        default:
          fprintf(stderr, "Protokollfehler: Falsches Ax-Telegramm\n");
      }
    } else if(tpuf[i].txt[0] == 'X' && tpuf[i].lg == 3)
      sp[tpuf[i].txt[2] >> 2].lebend = 0;
    else if(tpuf[i].lg != 1)
      fprintf(stderr, "Protokollfehler: Falsches Telegramm - A0\n");
  }
  printf("Spiel gestartet\n");

  /* Spiel selber */
  s_kehr = 1;
  while(zust != 5) {
    i = lies_telegramm(0);
    if(tpuf[i].txt[0] == 'P' && tpuf[i].lg % 3 == 1) {
      /* Positionstelegramm */
      j = tpuf[i].lg;
      while(j > 1) {
        j -= 3;
        s = tpuf[i].txt[j+2] >> 2;
        sp[s].x = tpuf[i].txt[j];
        sp[s].y = tpuf[i].txt[j+1];
        if(s == eig_s)
          eig_r = tpuf[i].txt[j+2] & 0x3;
        sp[s].lebend = 1;
        spielfeld[sp[s].x][sp[s].y] = 1;
      }
      /* Computergegnerzug ausfhren */
      if(zust == 3) {
        switch(strat) {
          case 1: /* primitiver Ausweicher */
            eig_rv = eig_r;
          break;
          case 2: /* Pltzlicher Abbieger */
            if(s_kehr == 0) {
              eig_rv = rand() & 3;
              s_kehr = rand() % 30 + 3;
            } else {
              eig_rv = eig_r;
              s_kehr--;
            }
          break;
          case 3: /* Langsamer Nachlufer */
            if(s_kehr == 0) {
              /* "Opfer" suchen */
              for(s = 0; s < n_spieler && (!sp[s].lebend || s == eig_s); s++);
              eig_rv = s == n_spieler ? rand() & 3 :
                abs(sp[s].x - sp[eig_s].x) > abs(sp[s].y - sp[eig_s].y)
              ?
                sp[s].x < sp[eig_s].x ? 2 : 0
              :
                sp[s].y < sp[eig_s].y ? 1 : 3 
              ;
              s_kehr = rand() % 20 + 3;
            } else {
              eig_rv = eig_r;
              s_kehr--;
            }
          break;            
          case 4: /* Hinterherjager */
            /* "Opfer" suchen */
            for(s = 0; s < n_spieler && (!sp[s].lebend || s == eig_s); s++);
            eig_rv = s == n_spieler ? eig_r :
              abs(sp[s].x - sp[eig_s].x) > abs(sp[s].y - sp[eig_s].y)
            ?
              sp[s].x < sp[eig_s].x ? 2 : 0
            :
              sp[s].y < sp[eig_s].y ? 1 : 3 
            ;
          break;
          case 5: /* Zufallstorkler */
            eig_rv = rand() & 3;
          break;
          case 6: /* Wandschleicher */
            eig_rv = eig_r - 1 & 0x3;
          break;
          default:
            fprintf(stderr, "Interner Fehler\n");
            exit(1);
        }
        /* "Schutzgott" :-) */
        for(i=0; i<4 && spielfeld[sp[eig_s].x+sx[eig_rv]]
                                 [sp[eig_s].y+sy[eig_rv]]; i++)
          eig_rv = eig_rv + 1 & 0x3;
        if(eig_rv != eig_r) {
          /* Nur im Falle von abweichender Richtung senden */
          r_tel = '@' + eig_rv;
          schreibe_telegramm(&r_tel, 1);
          /* eig_r wird von der Server-Rckantwort gesetzt! */
        }
      }
    } else if(tpuf[i].lg == 1) {
      if(tpuf[i].txt[0] >> 2 == eig_s)
        eig_r = tpuf[i].txt[0] & 0x3;
    } else if(tpuf[i].lg == 3 && (tpuf[i].txt[0] == 'K'
                               || tpuf[i].txt[0] == 'X')) {
      s = tpuf[i].txt[2] >> 2;
      sp[s].lebend = 0;
      if(s == eig_s) {
        printf("Autopilot gestrzt! :-(\n");
        zust = 4;
      } else
        printf("Spieler %d gestrzt\n", s+1);
    } else if(tpuf[i].lg > 2 && !strncmp(tpuf[i].txt, "Ra", 2))
      zust = 5;
    else
      fprintf(stderr, "Protokollfehler: Falsches Telegramm in Spiel\n");
  }
  printf("Spiel beendet.\n");
  schliesse();
  printf("Verbindung getrennt und Autopilot beendet.\n");
  return 0;
}
